@@ -1,7 +1,7 @@
 # -*- perl -*-
 
 #
-# $Id: Convert.pm,v 2.1 2003/02/07 11:38:53 eserte Exp $
+# $Id: Convert.pm,v 2.2 2003/05/29 22:02:45 eserte Exp $
 # Author: Slaven Rezic
 #
 # Copyright (C) 2001,2003 Slaven Rezic. All rights reserved.
@@ -16,13 +16,13 @@ package GD::Convert;
 
 use strict;
 use vars qw($VERSION $DEBUG);
-$VERSION = sprintf("%d.%02d", q$Revision: 2.1 $ =~ /(\d+)\.(\d+)/);
+$VERSION = sprintf("%d.%02d", q$Revision: 2.2 $ =~ /(\d+)\.(\d+)/);
 
 sub import {
     my($pkg, @args) = @_;
     foreach my $arg (@args) {
 	my($f, $as) = split /=/, $arg;
-	if ($f =~ /^(gif|newFromGif)$/) {
+	if ($f =~ /^(gif|newFromGif|newFromGifData)$/) {
 	    if ($as eq 'any') {
 		# check whether GD handles the gif itself
 		if ($GD::VERSION <= 1.19 ||
@@ -199,6 +199,57 @@ sub ppm {
     $ppm;
 }
 
+sub newFromPpmData {
+    my($self, $data, $truecolor) = @_;
+    (my $signature, my $dimensions, my $maxval, $data) = split /\n/, $data, 4;
+    if ($signature ne 'P6') {
+	die "Can handle only P6 (ppm raw) files";
+    }
+    my($width, $height) = split /\s+/, $dimensions;
+    if ($maxval != 255) {
+	die "Can handle only ppm files with maxval=255";
+    }
+    my $gd;
+    if ($GD::VERSION >= 2 && defined $truecolor) {
+	$gd = $self->new($width, $height, $truecolor);
+    } else {
+	$gd = $self->new($width, $height);
+    }
+    my %palette;
+    my $x = 0;
+    my $y = 0;
+    for(my $i = 0; $i < length($data); $i+=3) {
+	my($r,$g,$b) = map { ord } split //, substr($data, $i, 3);
+	my $color;
+	if (exists $palette{"$r/$g/$b"}) {
+	    $color = $palette{"$r/$g/$b"};
+	} else {
+	    $color = $gd->colorAllocate($r,$g,$b);
+	    $palette{"$r/$g/$b"} = $color;
+	}
+	$gd->setPixel($x, $y, $color);
+	$x++;
+	if ($x >= $width) {
+	    $x = 0;
+	    $y++;
+	    if ($y > $height) {
+		die "Image data does not match dimensions $width x $height";
+	    }
+	}
+    }
+
+    $gd;
+}
+
+sub newFromPpm {
+    my($self, $file, $truecolor) = @_;
+    open(FH, $file) or die "Can't open $file: $!";
+    local $/ = undef;
+    my $data = <FH>;
+    close FH;
+    $self->newFromPpmData($data, $truecolor);
+}
+
 sub _get_header {
     my $gdref = shift;
     my $is_gd2 = 0;
@@ -308,12 +359,93 @@ sub _gif_external {
 
 }
 
+sub _newFromGif_external {
+    my($self, $ext_type, $source_type, $data, $truecolor) = @_;
+
+    if ($source_type eq 'file') {
+	# $data is a file name
+	open(FH, $data) or die "Can't open $data: $!";
+	local $/ = undef;
+	$data = <FH>;
+	close FH;
+    }
+
+    my @cmd;
+    my $input_type;
+
+    if ($ext_type eq 'netpbm') {
+	@cmd = ("giftopnm");
+	$input_type = "pnm";
+    } elsif ($ext_type eq 'imagemagick') {
+	my $can_png;
+	if (GD::Image->can('png')) {
+	    # Prefer gif => png, because transparency information won't get
+	    # lost.
+	    $input_type = "png";
+	    $can_png = 1;
+	} else {
+	    $input_type = "pnm";
+	}
+
+	@cmd = ("convert");
+
+	if (!$can_png) {
+	    push @cmd, "gif:-", "ppm:-";
+	} else {
+	    push @cmd, "gif:-", "png:-";
+	}
+    } else {
+	die "Unhandled type $ext_type";
+    }
+
+    require IPC::Open3;
+
+    warn "Cmd: @cmd\n" if $GD::Convert::DEBUG;
+    my $pid = IPC::Open3::open3(\*WTR, \*RDR, \*ERR, @cmd);
+    die "Can't create process for @cmd" if !defined $pid;
+    binmode RDR;
+    binmode WTR;
+    print WTR $data;
+    close WTR;
+
+    my $data2;
+    {
+	local $/ = undef;
+	$data2 = scalar <RDR>;
+    }
+    close RDR;
+
+    my $cmd;
+    if ($input_type eq 'png') {
+	$cmd = "newFromPngData";
+    } else {
+	$cmd = "newFromPpmData";
+    }
+
+    my $gd;
+    if ($GD::VERSION >= 2 && defined $truecolor) {
+	$gd = $self->$cmd($data2, $truecolor);
+    } else {
+	$gd = $self->$cmd($data2);
+    }
+    $gd;
+}
+
 sub gif_netpbm      { shift->_gif_external("netpbm", @_) }
 sub gif_imagemagick { shift->_gif_external("imagemagick", @_) }
 
-#XXX implementation pending...
-sub newFromGif_netpbm      { die "NYI" }
-sub newFromGif_imagemagick { die "NYI" }
+sub newFromGif_netpbm      {
+    shift->_newFromGif_external("netpbm", "file", @_);
+}
+sub newFromGif_imagemagick {
+    shift->_newFromGif_external("imagemagick", "file", @_);
+}
+sub newFromGifData_netpbm      {
+    shift->_newFromGif_external("netpbm", "data", @_);
+}
+sub newFromGifData_imagemagick {
+    shift->_newFromGif_external("imagemagick", "data", @_);
+}
 
 #XXX merge with GD::Wbmp, delete GD::Wbmp
 sub _wbmp {
@@ -336,16 +468,48 @@ GD::Convert - additional output formats for GD
 =head1 SYNOPSIS
 
     use GD;
-    use GD::Convert 'gif=gif_netpbm';
+    use GD::Convert qw(gif=gif_netpbm newFromGif=newFromGif_imagemagick);
     ...
     $gd->ppm;
     $gd->xpm;
     $gd->gif;
+    ...
+    $gd = GD::Image->newFromPpmData(...);
+    $gd = GD::Image->newFromGif(...);
 
 =head1 DESCRIPTION
 
-This module provides additional output functions for the GD module:
-ppm, xpm and gif_netpbm/gif_imagemagick.
+This module provides additional output methods for the GD module:
+C<ppm>, C<xpm>, C<gif_netpbm> and C<gif_imagemagick>, and also
+additional constructors: C<newFromPpm>, C<newFromPpmData>,
+C<newFromGif_netpbm>, C<newFromGifData_netpbm>,
+C<newFromGif_imagemagick>, C<newFromGifData_imagemagick>.
+
+The new methods go into the C<GD> namespace.
+
+For convenience, it is possible to set shorter names for the C<gif>
+etc. methods:
+
+=over 4
+
+=item gif=gif_netpbm
+
+Use external commands from netpbm to create GIF images.
+
+=item gif=gif_imagemagick
+
+Use external commands from imagemagick to create GIF images.
+
+=item gif=any
+
+Use any of the above methods to create GIF images.
+
+=back
+
+The same convenience importer is defined for C<newFromGif> and
+C<newFromGifData>.
+
+The new methods and constructors:
 
 =over 4
 
@@ -374,6 +538,34 @@ normal color.
 This is the same as C<gif_netpbm>, instead it is using the C<convert>
 program of ImageMagick.
 
+=item $image = GD::Image->newFromPpm($file, [$truecolor])
+
+Create a GD image from the named ppm file, Only raw ppm files
+(signature P6) are supported.
+
+=item $image = GD::Image->newFromPpmData($data, [$truecolor])
+
+Create a GD image from the data string containing ppm data. Only raw
+ppm files are supported.
+
+=item $image = GD::Image->newFromGif_netpbm($file, [$truecolor]);
+
+Create a GD image from the named file using external netpbm programs.
+
+=item $image = GD::Image->newFromGifData_netpbm($file, [$truecolor]);
+
+Create a GD image from the data string using external netpbm programs.
+
+=item $image = GD::Image->newFromGif_imagemagick($file, [$truecolor]);
+
+Create a GD image from the named file using external ImageMagick
+programs.
+
+=item $image = GD::Image->newFromGifData_imagemagick($file, [$truecolor]);
+
+Create a GD image from the data string using external ImageMagick
+programs.
+
 =back
 
 =head1 BUGS
@@ -398,5 +590,5 @@ it under the same terms as Perl itself.
 
 =head1 SEE ALSO
 
-GD(3), netpbm(1).
+L<GD>, L<netpbm(1)>, L<convert(1)>.
 
